XPerl_MainTanks = {}
local MainTankCount = 0
local MainTanks = {}
local Events = {}
local XUnits = {}
local UpdateTime = 0
local XGaps = 0
local XTitle
local pendingTankListChange		-- If in combat when tank list changes, then we'll defer it till next time we're out of combat

-- Dup colours used to show which tanks have the same target
local MTdupColours = {	{r = "0.8", g = "0.2", b = "0.2"},
			{r = "0.2", g = "0.2", b = "0.8"},
			{r = "0.8", g = "0.2", b = "0.8"},
			{r = "0.2", g = "0.8", b = "0.2"},
			{r = "0.2", g = "0.8", b = "0.8"}}

local GAP_SPACING = 10
XPERL_UNIT_HEIGHT_MIN = 17
XPERL_UNIT_HEIGHT_MAX = 30
XPERL_UNIT_WIDTH_MIN = 50
XPERL_UNIT_WIDTH_MAX = 125

if (not XPerlColourTable) then
	XPerlColourTable = setmetatable({},{
		__index = function(self, class)
			local c = RAID_CLASS_COLORS[class] or {r = 0, g = 0, b = 0}
			return format("|c00%02X%02X%02X", 255 * c.r, 255 * c.g, 255 * c.b)
		end
	})
end

-- Although your own "target" is updated immediately, any reference to ourself via our raid id is updated via the server
-- So, to cure the slow response times, we check if anything is 'me'. Only direct targets are updated immedately,
-- so we don't bother with target's target onwards.
local function SpecialCaseUnit(self)
	local id
	if (self.type == "MTT") then
		id = SecureButton_GetUnit(self:GetParent())
		if (id and UnitIsUnit("player", id)) then
			return "target"
		end
	elseif (self.type == "MT") then
		id = SecureButton_GetUnit(self)
		if (id and UnitIsUnit("player", id)) then
			return "player"
		end
		return id
	end

	return SecureButton_GetUnit(self)
end

----------------------------
------- DISPLAY ------------
----------------------------
local function UpdateUnit(self,forcedUpdate)

	if (self.hidden) then
		return
	end

	if (forcedUpdate) then
		local time = GetTime()
		if (self.update and time < self.update + 0.2) then
			-- Since we catch UNIT_HEALTH changes, we don't need to update always
			return
		end
	end

	self.update = GetTime()

	local xunit = SpecialCaseUnit(self)
	if (not xunit) then
		return
	end

	local name = UnitName(xunit)
	local percBar

	self.healthBar.bg:Show()

	if (name and name ~= UNKNOWNOBJECT) then
		-- Name
		self.text:SetText(name)
		self.text:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", 0, -12)

		local remCount = 1
		while ((self.text:GetStringWidth() >= (self:GetWidth() - 2)) and (string.len(name) > remCount)) do
			name = string.sub(name, 1, string.len(name) - remCount)..".."
			remCount = 3
			self.text:SetText(name)
		end

		if (UnitPlayerControlled(xunit)) then
			XPerl_ColourFriendlyUnit(self.text, xunit)
		else
			XPerl_SetUnitNameColor(xunit, self.text)
		end

		-- Health
		local healthMax = UnitHealthMax(xunit)
		local health = UnitHealth(xunit)
		percBar = health / healthMax
		local perc = percBar * 100

		self.healthBar:SetMinMaxValues(0, healthMax)
		self.healthBar:SetValue(health)

		if (UnitIsDead(xunit)) then
			self.healthBar.text:SetText(XPERL_LOC_DEAD)
		else
			if (XPerlConfigHelper.HealerMode == 1 and UnitInRaid(xunit)) then
				if (XPerlConfigHelper.HealerModeType == 1) then
					self.healthBar.text:SetText(health - healthMax.."/"..healthMax)
				else
					self.healthBar.text:SetText(health - healthMax)
				end
			else
				self.healthBar.text:SetText(floor(perc + 0.5).."%")
			end
		end
		if (XPerlConfigHelper.UnitHeight < 23) then
			self.healthBar.text:Hide()
		else
			self.healthBar.text:Show()
		end

		if (self.type == "MTT") then
			if (BigWigsTargetMonitor) then
				BigWigsTargetMonitor:TargetCheck(xunit)
			end

			local index = GetRaidTargetIndex(xunit)
			if (index) then
				SetRaidTargetIconTexture(self.raidIcon, index)
				self.raidIcon:Show()
			else
				self.raidIcon:Hide()
			end
		else
			self.raidIcon:Hide()
		end

		if (UnitAffectingCombat(xunit)) then
			self.combatIcon:Show()
		else
			self.combatIcon:Hide()
		end

		if (UnitIsCharmed(xunit)) then
			self.warningIcon:Show()
		else
			self.warningIcon:Hide()
		end
	else
		self.text:SetPoint("BOTTOMRIGHT", -2, 1)

		if (self.type == "MTT") then
			local id = SecureButton_GetUnit(self:GetParent())
			name = UnitName(id)

			if (name) then
				XPerl_ColourFriendlyUnit(self.text, id)
				self.text:SetText(format(XPERL_XS_TARGET, name))
				self.healthBar.bg:Hide()
			end
		else
			self.text:SetText(XPERL_NO_TARGET)
			self.text:SetTextColor(0.5, 0.5, 0.5, 1)
		end
		self.healthBar:SetMinMaxValues(0, 1)
		self.healthBar:SetValue(0)
		percBar = 0
		self.healthBar.text:Hide()
		self.raidIcon:Hide()
		self.combatIcon:Hide()
		self.warningIcon:Hide()
	end

	if (XPerl_UpdateSpellRange and XPerlConfig.RangeFinder.enabled == 1) then
		XPerl_UpdateSpellRange(self)
	else
		self:SetAlpha(1)
	end

	XPerl_SetSmoothBarColor (self.healthBar, percBar)
end

-- XPerl_RaidHelp_Show
function XPerl_RaidHelp_Show()
	XPerlConfigHelper.RaidHelper = 1
	XPerl_RaidHelperCheck:Show()		-- XPerl_EnableDisable()
	return true
end

-- XPerl_MTListUnit_OnEnter
function XPerl_MTListUnit_OnEnter(self)

	local x = XPerl_Frame:GetCenter()
	local a1, a2 = "ANCHOR_LEFT", "ANCHOR_TOPLEFT"
	if ( x < (GetScreenWidth() / 2)) then
		a1, a2 = "ANCHOR_RIGHT", "ANCHOR_TOPRIGHT"
	end

	if (XPerlConfigHelper.Tooltips == 1) then
		local partyid = SecureButton_GetUnit(self)

		if (partyid and UnitName(partyid) and XPerlConfigHelper.TooltipsWhich == 2) then
			GameTooltip:SetOwner(self, a1)
			GameTooltip:SetUnit(partyid)

			if (self.type == "MTT" and XPerlConfigHelper.TooltipsWhich == 2) then
				local parentID = SecureButton_GetUnit(self:GetParent())
				if (parentID) then
					XPerl_BottomTip:SetOwner(GameTooltip, a2, 0, 10)
					XPerl_BottomTip:SetUnit(parentID)
					XPerl_BottomTip:SetBackdropColor(0.1, 0.4, 0.1, 0.75)
				end
			end
		else
			if (XPerlConfigHelper.TooltipsWhich == 0) then
				GameTooltip:SetOwner(self, a1)
				GameTooltip:SetUnit(partyid)
			else
				if (self.type == "MTT") then
					partyid = SecureButton_GetUnit(self:GetParent())
				end

				GameTooltip:SetOwner(self, a1)
				GameTooltip:SetUnit(partyid)
				GameTooltip:SetBackdropColor(0.1, 0.4, 0.1, 0.75)
			end
		end
	end
end

-- XPerl_SetupFrameSimple
function XPerl_SetupFrameSimple(argFrame, alpha)
	argFrame:SetBackdropBorderColor(0.5, 0.5, 0.5, alpha or 0.8)
	argFrame:SetBackdropColor(0, 0, 0, alpha or 0.8)
end

-- XPerl_HelperBarTextures
if (XPerl_GetBarTexture) then
function XPerl_HelperBarTextures(tex)
	local function UnitTex(v, tex)
		v.healthBar.tex:SetTexture(tex)
		if (XPerlConfig and XPerlConfig.BackgroundTextures == 1) then
			v.healthBar.bg:SetTexture(tex)
		else
			v.healthBar.bg:SetTexture("Interface\\TargetingFrame\\UI-StatusBar")
		end
	end

	for k,v in pairs(XUnits) do
		UnitTex(v, tex)
	end
end
end

-- RemoveDupMT
local function RemoveDupMT(name)
	for k, v in pairs(XPerl_MainTanks) do
		if (v[2] == name) then
			XPerl_MainTanks[k] = nil
			return true
		end
	end
end

-- GetRaidIDByName
local function GetRaidIDByName(name)
	for i = 1, GetNumRaidMembers() do
		if (UnitName("raid"..i) == name) then
			return i
		end
	end
end

-- GetUnitRank
local function GetUnitRank(name)
	local index = GetRaidIDByName(name)
	if (index) then
		local _, rank = GetRaidRosterInfo(index)
		return rank
	end
	return 0
end

-- ValidateTankList
-- Check the roster for any tanks that have left the raid
local function ValidateTankList()
	for index,entry in pairs(XPerl_MainTanks) do
		local found
		for i = 1,GetNumRaidMembers() do
			if (UnitName("raid"..i) == entry[2]) then
				found = true
				break
			end
		end

		if (not found) then
			XPerl_MainTanks[index] = nil
		end
	end
end

-- XPerl_MTRosterChanged()
function XPerl_MTRosterChanged()
	ValidateTankList()

	MainTankCount = 0
	for i in pairs(XPerl_MainTanks) do
		MainTankCount = MainTankCount + 1
	end

	if (MainTankCount == 0 or XPerlConfigHelper.UseCTRATargets == 0) then
		-- If no defined tanks, then make a list from the warriors in raid
		MainTanks = {}
		MainTankCount = 0

		if (XPerlConfigHelper.NoAutoList == 0) then
			for i = 1,GetNumRaidMembers() do
				local _, class = UnitClass("raid"..i)
				if (class == "WARRIOR") then
					tinsert(MainTanks, {i, UnitName("raid"..i)})
					MainTankCount = MainTankCount + 1
				end
			end
		end
	else
		MainTankCount = 0
		for index,entry in pairs(XPerl_MainTanks) do
			local raidid = GetRaidIDByName(entry[2])
			if (raidid) then
				MainTankCount = MainTankCount + 1
				entry[1] = raidid
			end
		end

		MainTanks = XPerl_MainTanks
	end

	XPerl_MakeTankList()
end

-- CT_RA_IsSendingWithVersion override
-- Override this so that we get the MT list if there's only us in the raid with a CTRA version registered
local old_CT_RA_IsSendingWithVersion
if (CT_RA_IsSendingWithVersion) then
	if (CT_RA_IsSendingWithVersion ~= XPerl_IsSendingWithVersion) then
		old_CT_RA_IsSendingWithVersion = CT_RA_IsSendingWithVersion
		CT_RA_IsSendingWithVersion = function(version)
			if (version == 1.08) then
				return 1
			end
			return old_CT_RA_IsSendingWithVersion(version)
		end
	end
end

-- ProcessCTRAMessage
XPerlCTRASet = {}
local function ProcessCTRAMessage(unitName, msg)

	local mtListUpdate

	if (strsub(msg, 1, 4) == "SET ") then
		if (GetUnitRank(unitName) < 1) then
			return
		end

		local num, name = strmatch(msg, "^SET (%d+) (.+)$")
		if (num and name) then
			num = tonumber(num)
			local mtID = 0
			for i = 1, GetNumRaidMembers() do
				if (UnitName("raid"..i) == name) then
					mtID = i
					break
				end
			end

			if (XPerl_MainTanks[num]) then
				if (XPerl_MainTanks[num][1] == mtID and XPerl_MainTanks[num][2] == name) then
					return			-- No Change
				end
			end

			RemoveDupMT(name)

			XPerl_MainTanks[tonumber(num)] = {mtID, name}

			mtListUpdate = true
		end

	elseif (strsub(msg, 1, 2) == "R ") then
		if (GetUnitRank(unitName) < 1) then
			return
		end

		local name = strmatch(msg, "^R (.+)$")
		if (name) then
			mtListUpdate = RemoveDupMT(name)
		end
	end

	if (mtListUpdate) then
		-- XPerl_MTRosterChanged()
		XPerl_RaidHelperCheck:Show()		-- XPerl_EnableDisable()
	end
end

-- XPerl_SplitCTRAMessage
function XPerl_SplitCTRAMessage(msg, char)
	local arr = { }
	while (strfind(msg, char) ) do
		local iStart, iEnd = strfind(msg, char)
		tinsert(arr, strsub(msg, 1, iStart-1))
		msg = strsub(msg, iEnd+1, strlen(msg))
	end
	if ( strlen(msg) > 0 ) then
		tinsert(arr, msg)
	end
	return arr
end

-- XPerl_ParseCTRA
function XPerl_ParseCTRA(nick, msg, func)
	if (strfind(msg, "#")) then
		local arr = XPerl_SplitCTRAMessage(msg, "#")
		for i,subMsg in pairs(arr) do
			func(nick, subMsg)
		end
	else
		func(nick, msg)
	end
end

-- CHAT_MSG_ADDON
function Events:CHAT_MSG_ADDON()
	if (arg1 == "CTRA" and arg3 == "RAID") then
		XPerl_ParseCTRA(arg4, arg2, ProcessCTRAMessage)
	end
end

----------------------------
---------- EVENTS ----------
----------------------------

-- OnEvent
function XPerl_OnEvent()
	local Event = Events[event]
	if (Event) then
		Event()
	else
XPerl_ShowMessage("EXTRA EVENT")
	end
end

-- XPerl_GetUnit
local function XPerl_GetUnit(unit)
	for i,search in pairs(XUnits) do
		if (search.type == "MTT") then
			local partyid = SecureButton_GetUnit(search:GetParent())
			if (partyid == unit) then
				return search
			end
		end
	end
end

-- ScanForMTDups
-- Check MTs all have unique targets
local function ScanForMTDups()
	local dup = 1
	local dups = {}

	for i,unit in pairs(XUnits) do
		XPerl_SetupFrameSimple(unit, XPerlConfigHelper.Background_Transparency)
	end

	for i,t1 in pairs(MainTanks) do
		local any = false

		if (UnitName("raid"..t1[1].."target") and UnitName("raid"..t1[1].."target") ~= UNKNOWNOBJECT and not UnitIsDeadOrGhost("raid"..t1[1])) then
			for j,t2 in pairs(MainTanks) do
				if (j > i and not dups["raid"..t2[1]]) then
					if (UnitIsUnit("raid"..t1[1].."target", "raid"..t2[1].."target")) then
						dups["raid"..t2[1]] = 1

						local unit1 = XPerl_GetUnit("raid"..t1[1])
						local unit2 = XPerl_GetUnit("raid"..t2[1])

						if (unit1) then
							unit1:SetBackdropBorderColor(MTdupColours[dup].r, MTdupColours[dup].g, MTdupColours[dup].b)
							unit1:SetBackdropColor(MTdupColours[dup].r, MTdupColours[dup].g, MTdupColours[dup].b)
						end
						if (unit2) then
							unit2:SetBackdropBorderColor(MTdupColours[dup].r, MTdupColours[dup].g, MTdupColours[dup].b)
							unit2:SetBackdropColor(MTdupColours[dup].r, MTdupColours[dup].g, MTdupColours[dup].b)
						end
						any = true
					end
				end
			end

			if (any and dup < 5) then
				dup = dup + 1
			end
		end
	end
end

-- MoveArrow
local function MoveArrow(unit)
	XPerl_MyTarget:SetParent(unit)
	XPerl_MyTarget:SetPoint("RIGHT", unit, "LEFT", 2 - (XPerlConfigHelper.MTLabels * 10), 0)
	XPerl_MyTarget:Show()
end

-- OnUpdate
local function OnUpdate(self, elapsed)
	UpdateTime = UpdateTime + elapsed
	if (UpdateTime > 0.5) then
		-- Forced update here for anything we don't receive events for
		UpdateTime = 0
		for k,unit in pairs(XUnits) do
			if (unit.type == "MTT" or unit.type == "MTTT") then
				UpdateUnit(unit, true)
			end
		end
	end
end

-- VARIABLES_LOADED
function Events:VARIABLES_LOADED()
	this:UnregisterEvent(event)

	if (type(XPerl_MainTanks) ~= "table") then
		XPerl_MainTanks = {}
	end

	XPerl_MTTargets:UnregisterEvent("UNIT_NAME_UPDATE")	-- Fix for WoW 2.1 UNIT_NAME_UPDATE issue
	XPerl_MTTargets:SetAttribute("template", "XPerl_MTList_UnitTemplate")

	XPerl_Startup()
	XPerl_RaidHelperCheck:Show()		-- XPerl_EnableDisable()
end

-- Registration
local function Registration()
	local list = {	"UNIT_HEALTH",
			"UNIT_MAXHEALTH",
			"UNIT_TARGET",
			"PLAYER_TARGET_CHANGED",
			"PLAYER_REGEN_ENABLED",
			"PLAYER_REGEN_DISABLED",
			"PLAYER_ENTERING_WORLD",
			"CHAT_MSG_ADDON"}

	for i,a in pairs(list) do
		if (GetNumRaidMembers() > 0 and XPerlConfigHelper.RaidHelper == 1) then
			XPerl_Frame:RegisterEvent(a)
		else
			XPerl_Frame:UnregisterEvent(a)
		end
	end
end

-- XPerl_HelperCheck_OnUpdate
function XPerl_HelperCheck_OnUpdate(self)
	self:Hide()
	XPerl_EnableDisable()
end

-- XPerl_EnableDisable
function XPerl_EnableDisable()
	if (InCombatLockdown()) then
		pendingTankListChange = true
		return
	end

	if (XPerlConfigHelper and XPerlConfigHelper.RaidHelper == 1 and GetNumRaidMembers() > 0) then
		if (not XPerl_Frame:IsShown()) then
			XPerl_Frame:Show()
			XPerl_Frame:SetScript("OnUpdate", OnUpdate)
		end

		if (CT_RAMenu_Options and CT_RA_UpdateVisibility) then
			XPerlConfigHelper.OldCTMTListHide = CT_RAMenu_Options["temp"]["HideMTs"]
			CT_RAMenu_Options["temp"]["HideMTs"] = true
			CT_RA_UpdateVisibility(1)
		end
	else
		if (XPerl_Frame:IsShown()) then
			XPerl_Frame:Hide()
			XPerl_Frame:SetScript("OnUpdate", nil)
		end

		if (XPerlConfigHelper and XPerlConfigHelper.RaidHelper == 0 and CT_RAMenu_Options and CT_RA_UpdateVisibility) then
			if (XPerlConfigHelper.OldCTMTListHide ~= true) then
				XPerlConfigHelper.OldCTMTListHide = nil
				CT_RAMenu_Options["temp"]["HideMTs"] = false
				CT_RA_UpdateVisibility(1)
			end
		end
	end

	Registration()
	XPerl_MTRosterChanged()
end

-- RAID_ROSTER_UPDATE
function Events:RAID_ROSTER_UPDATE()
	if (GetNumRaidMembers() == 0) then
		MainTanks = {}
	end

	XPerl_RaidHelperCheck:Show()		-- XPerl_EnableDisable()
end

-- UNIT_HEALTH
function Events:UNIT_HEALTH()
	if (strmatch(arg1, "^raid") and not strmatch(arg1, "pet")) then
		for k,unit in pairs(XUnits) do
			local partyid = SecureButton_GetUnit(unit)
			if (partyid and UnitIsUnit(partyid, arg1)) then
				UpdateUnit(unit)
			end
		end
	end
end

Events.UNIT_MAXHEALTH = Events.UNIT_HEALTH

-- CheckArrowPosition
local function CheckArrowPosition()
	local arrowDone
	local arrowType = "MTT"
	local arrowExtra = ""
	if (XPerlConfigHelper.ShowMT == 1) then
		arrowType = "MT"
		arrowExtra = "target"
	end

	XPerl_MyTarget:Hide()
	XPerl_MainAssist:Hide()
	XPerl_MainTank:Hide()

	local maAssignment, tankAssignment
	if (GetPartyAssignment) then
		maAssignment = GetPartyAssignment("mainassist")
		tankAssignment = GetPartyAssignment("maintank")
	end

	local maDone, mtDone
	for i,unit in pairs(XUnits) do
		local xunit = SpecialCaseUnit(unit)
		if (xunit) then
			if (not arrowDone and unit.type == arrowType and UnitIsUnit("target", xunit..arrowExtra) and UnitExists("target")) then
				MoveArrow(unit)
				arrowDone = true
			end

			if (unit.type == "MTT") then
				xunit = SpecialCaseUnit(unit:GetParent())
			end

			local name = UnitName(xunit)
			if (maAssignment and unit.type == arrowType and not maDone and name == maAssignment) then
				maDone = true
				XPerl_MainAssist:SetParent(unit.healthBar)
				XPerl_MainAssist:ClearAllPoints()

				if (XPerlConfigHelper.MTLabels == 1) then
					XPerl_MainAssist:SetPoint("BOTTOMRIGHT", unit, "BOTTOMLEFT", 3, 1)
				else
					XPerl_MainAssist:SetPoint("BOTTOMRIGHT", -1, 1)
				end
				XPerl_MainAssist:Show()
			end

			if (tankAssignment and unit.type == arrowType and not mtDone and name == tankAssignment) then
				mtDone = true
				XPerl_MainTank:SetParent(unit.healthBar)
				XPerl_MainTank:ClearAllPoints()

				if (XPerlConfigHelper.MTLabels == 1) then
					XPerl_MainTank:SetPoint("BOTTOMRIGHT", unit, "BOTTOMLEFT", 3, 1)
				else
					XPerl_MainTank:SetPoint("BOTTOMRIGHT", -1, 1)
				end
				XPerl_MainTank:Show()
			end
		end
	end
end

-- TargetChanged
local function TargetChanged(xunit)
	local unit, i = XPerl_MTTargets:GetAttribute("child1"), 1
	while (unit) do
		local id = unit:GetAttribute("unit")
		if (id and UnitIsUnit(id, xunit)) then
			UpdateUnit(unit.target)
			UpdateUnit(unit.targettarget)
			any = true
			break
		end

		i = i + 1
		unit = XPerl_MTTargets:GetAttribute("child"..i)
	end

	if (any) then
		ScanForMTDups()
		CheckArrowPosition()
	end
end

-- PLAYER_TARGET_CHANGED
function Events:PLAYER_TARGET_CHANGED()
	if (XPerl_Frame:IsShown()) then
		TargetChanged("player")
	end
end

-- UNIT_TARGET
function Events:UNIT_TARGET()
	if (XPerl_Frame:IsShown() and strmatch(arg1, "^raid")) then
		TargetChanged(arg1)
	end
end

-- DoButtons
local function DoButtons(disable)
	if (disable or InCombatLockdown()) then
		XPerl_Frame_CloseButton:Disable()
		XPerl_Frame_ToggleShowMT:Disable()
		XPerl_Frame_ToggleMTTargets:Disable()
		XPerl_Frame_ToggleLabels:Disable()
	else
		XPerl_Frame_CloseButton:Enable()
		XPerl_Frame_ToggleShowMT:Enable()
		XPerl_Frame_ToggleMTTargets:Enable()
		XPerl_Frame_ToggleLabels:Enable()
	end

	XPerl_Frame_ToggleShowMT:SetButtonTex()
	XPerl_Frame_ToggleMTTargets:SetButtonTex()
	XPerl_Frame_ToggleLabels:SetButtonTex()
end

-- PLAYER_REGEN_ENABLED
function Events:PLAYER_REGEN_ENABLED()
	if (pendingTankListChange) then
		XPerl_MakeTankList()
	end
	DoButtons()
end
Events.PLAYER_ENTERING_WORLD = Events.PLAYER_REGEN_ENABLED

-- PLAYER_REGEN_DISABLED
function Events:PLAYER_REGEN_DISABLED()
	DoButtons(true)

end

-- The number of 'Other' colums that can be shown
local function DisplayableColumns()
	return 1 + XPerlConfigHelper.ShowMT + XPerlConfigHelper.MTTargetTargets
end

-- XPerl_SetTitle
function XPerl_SetTitle()
	if (DisplayableColumns() > 1 and XPerlConfigHelper.UnitWidth >= 70) then
		if (XPerl_MainTanks == MainTanks) then
			XPerl_Frame_Title:SetText(XPERL_TITLE_MT_LONG)
		else
			XPerl_Frame_Title:SetText(XPERL_TITLE_WARRIOR_LONG)
		end
	else
		if (XPerl_MainTanks == MainTanks) then
			XPerl_Frame_Title:SetText(XPERL_TITLE_MT_SHORT)
		else
			XPerl_Frame_Title:SetText(XPERL_TITLE_WARRIOR_SHORT)
		end
	end
end

-- SetVisibility
local function SetVisibility()

	local left = 3 + (XPerlConfigHelper.MTLabels * 10)
	XPerl_Frame:ClearAllPoints()
	XPerl_Frame_TitleBar:ClearAllPoints()
	XPerl_MTTargets:ClearAllPoints()
	if (XPerlConfigHelper.MTListUpward == 1) then
		XPerl_Frame:SetPoint("BOTTOMLEFT", XPerl_MTList_Anchor, "BOTTOMLEFT", 0, 0)
		XPerl_Frame_TitleBar:SetPoint("BOTTOMLEFT", 3, 3)
		XPerl_Frame_TitleBar:SetPoint("TOPRIGHT", XPerl_Frame, "BOTTOMRIGHT", -3, 13)
		XPerl_MTTargets:SetAttribute("point", "BOTTOM")
		XPerl_MTTargets:SetPoint("BOTTOMLEFT", left, 13)
	else
		XPerl_Frame:SetPoint("TOPLEFT", XPerl_MTList_Anchor, "TOPLEFT", 0, 0)
		XPerl_Frame_TitleBar:SetPoint("TOPLEFT", 3, -3)
		XPerl_Frame_TitleBar:SetPoint("BOTTOMRIGHT", XPerl_Frame, "TOPRIGHT", -3, -13)
		XPerl_MTTargets:SetAttribute("point", "TOP")
		XPerl_MTTargets:SetPoint("TOPLEFT", left, -13)
	end

	local v, k = XPerl_MTTargets:GetAttribute("child1"), 1
	while (v) do
		-- We don't actually hide the MT if it's not shown, just make it unclickable and no visible content
		if (XPerlConfigHelper.ShowMT == 1) then
			v.hidden = false
			v.text:Show()
			v.healthBar:Show()
			v.healthBar.text:Show()
			v:EnableMouse(true)

			v:SetBackdrop({bgFile = "Interface\\AddOns\\XPerl_RaidHelper\\images\\XPerl_FrameBack", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 9, insets = { left = 3, right = 3, top = 3, bottom = 3 } } );

			v.target:SetPoint("TOPLEFT", v, "TOPRIGHT", 0, 0)
		else
			v.hidden = true
			v.text:Hide()
			v.healthBar:Hide()
			v.healthBar.text:Hide()
			v.combatIcon:Hide()
			v.warningIcon:Hide()
			v.raidIcon:Hide()
			v:EnableMouse(false)

			v:SetBackdrop({bgFile = "Interface\\AddOns\\XPerl_RaidHelper\\images\\XPerl_FrameBack", edgeFile = "", tile = true, tileSize = 16, edgeSize = 9, insets = { left = 3, right = 3, top = 3, bottom = 3 } } );

			v.target:SetPoint("TOPLEFT", v, "TOPLEFT", 0, 0)
		end

		UpdateUnit(v)

		if (XPerlConfigHelper.MTTargetTargets == 1) then
			v.targettarget:Show()
		else
			v.targettarget:Hide()
		end

		local label = v.label
		if (XPerlConfigHelper.MTLabels == 1) then
			local u = v:GetAttribute("unit")
			if (u) then
				local lbl
				local name = UnitName(u)
				for i,tank in pairs(MainTanks) do
					if (tank[2] == name) then
						lbl = i
						break
					end
				end

				label:SetText(lbl)
				label:Show()
			end
		else
			label:Hide()
		end

		k = k + 1
		v = XPerl_MTTargets:GetAttribute("child"..k)
	end
end

-- XPerl_MakeTankList
function XPerl_MakeTankList()
	if (InCombatLockdown()) then
		pendingTankListChange = true
		return
	else
		pendingTankListChange = nil
	end

	if (not XPerlConfigHelper or XPerlConfigHelper.RaidHelper == 0 or GetNumRaidMembers() == 0) then
		XPerl_SetFrameSizes()
		return
	end

	XPerl_SetTitle()
	XPerl_MyTarget:Hide()
	XGaps = 0

	local tanks = {}
	for i = 1,10 do
		if (MainTanks[i]) then
			tinsert(tanks, MainTanks[i][2])
			if (#tanks >= XPerlConfigHelper.MaxMainTanks) then
				break
			end
		end
	end

	if (#tanks == 0) then
		XPerl_Frame:Hide()
	else
		XPerl_Frame:Show()
		XPerl_Tutorial(12)
	end

	XPerl_MTTargets:SetAttribute("nameList", table.concat(tanks, ","))
	XPerl_MTTargets:Show()

	if (XPerl_NoFadeBars) then XPerl_NoFadeBars() end

	XPerl_SetFrameSizes()
end

-- SetUnitSettings
local function SetUnitSettings(self)
	self:SetWidth(XPerlConfigHelper.UnitWidth)
	self:SetHeight(XPerlConfigHelper.UnitHeight)

	if (XPerl_GetBarTexture) then
		local tex = XPerl_GetBarTexture()
		self.healthBar.tex:SetTexture(tex)

		if (XPerlConfig.BackgroundTextures == 1) then
			self.healthBar.bg:SetTexture(tex)
		else
			self.healthBar.bg:SetTexture("Interface\\TargetingFrame\\UI-StatusBar")
		end
	end

	self.healthBar:ClearAllPoints()
	if (XPerlConfigHelper.UnitHeight > 16) then
		self.healthBar:SetPoint("TOPLEFT", 3, -14)
		self.healthBar:SetPoint("BOTTOMRIGHT", -3, 2)
	else
		self.healthBar:SetPoint("TOPLEFT", 4, -4)
		self.healthBar:SetPoint("BOTTOMRIGHT", -4, 4)
	end

	if (XPerl_RegisterHighlight) then
		XPerl_RegisterHighlight(self, 3)
	else
		local tex = self:GetHighlightTexture()
		tex:SetVertexColor(0.86, 0.82, 0.41)
	end
end

-- onAttrChanged
local function onAttrChanged(self, name, value)
	if (name == "unit") then
		if (value) then
			UpdateUnit(self)
			UpdateUnit(self.target)
			UpdateUnit(self.targettarget)
		end
	end
end

-- XPerl_MTList_ChildUnits
function XPerl_MTList_ChildUnits(self)
	self.type = "MT"
	self.hidden = false
	self:SetAttribute("*type1", "target")
	XPerl_RegisterClickCastFrame(self)
	SetUnitSettings(self)

	local c1 = CreateFrame("Button", self:GetName().."Target", self, "XPerl_MTList_UnitTemplate")
	c1.type = "MTT"
	c1:SetAttribute("useparent-unit", true)
	c1:SetAttribute("unitsuffix", "target")
	c1:SetAttribute("*type1", "target")
	XPerl_RegisterClickCastFrame(c1)

	c1:SetPoint("TOPLEFT", self, "TOPRIGHT", 0, 0)
	SetUnitSettings(c1)

	local c2 = CreateFrame("Button", self:GetName().."TargetTarget", self, "XPerl_MTList_UnitTemplate")
	c2.type = "MTTT"
	c2:SetAttribute("useparent-unit", true)
	c2:SetAttribute("unitsuffix", "targettarget")
	c2:SetAttribute("*type1", "target")
	XPerl_RegisterClickCastFrame(c2)

	c2:SetPoint("TOPLEFT", c1, "TOPRIGHT", 0, 0)
	SetUnitSettings(c2)

	self.target = c1
	self.targettarget = c2

	self:SetScript("OnShow", UpdateUnit)
	c1:SetScript("OnShow", UpdateUnit)
	c2:SetScript("OnShow", UpdateUnit)

	self:SetScript("OnAttributeChanged", onAttrChanged)

	self:Show()
	c1:Show()
	c2:Show()

	tinsert(XUnits, self)
	tinsert(XUnits, c1)
	tinsert(XUnits, c2)
end

-- XPerl_SetFrameSizes
function XPerl_SetFrameSizes()

	local tanks = MainTankCount

	SetVisibility()		-- Change which of MT, MTT and MTTT we can see
	ScanForMTDups()

	if (tanks > XPerlConfigHelper.MaxMainTanks) then
		tanks = XPerlConfigHelper.MaxMainTanks
	end

	XPerl_MTTargets:SetHeight((XPerlConfigHelper.UnitHeight * tanks) + (XGaps * GAP_SPACING))

	local ExtraWidth = (XPerlConfigHelper.MTLabels * 10);	-- 0 for off, 10 for on
	XLastColumCount = DisplayableColumns()
	XPerl_Frame:SetWidth(XLastColumCount * XPerlConfigHelper.UnitWidth + 6 + ExtraWidth)

	XPerl_Frame:SetHeight((XPerlConfigHelper.UnitHeight * tanks) + (XGaps * GAP_SPACING) + 6 + XPerl_Frame_TitleBar:GetHeight())

	CheckArrowPosition()
end

-- XPerl_MTList_ApplyUnitSizes
function XPerl_MTList_ApplyUnitSizes()
	for k,unit in pairs(XUnits) do
		unit:SetWidth(XPerlConfigHelper.UnitWidth)
		unit:SetHeight(XPerlConfigHelper.UnitHeight)
		UpdateUnit(unit)
	end
end

-- XPerl_Toggle_UseCTRATargets
function XPerl_Toggle_UseCTRATargets()
	if (XPerlConfigHelper.UseCTRATargets == 1) then
		XPerlConfigHelper.UseCTRATargets = 0
	else
		XPerlConfigHelper.UseCTRATargets = 1
	end

	--XPerl_MTRosterChanged()
	XPerl_RaidHelperCheck:Show()		-- XPerl_EnableDisable()
	return true
end

function XPerl_Toggle_MTTargets(noChange)
	if (not noChange) then
		if (XPerlConfigHelper.MTTargetTargets == 1) then
			XPerlConfigHelper.MTTargetTargets = 0
		else
			XPerlConfigHelper.MTTargetTargets = 1
		end
	end

	XPerl_SetFrameSizes()

	XPerl_Frame_ToggleMTTargets:SetButtonTex()
	return true
end

-- XPerl_Toggle_ToggleLabels
function XPerl_Toggle_ToggleLabels(noChange)

	if (not noChange) then
		if (XPerlConfigHelper.MTLabels == 1) then
			XPerlConfigHelper.MTLabels = 0
		else
			XPerlConfigHelper.MTLabels = 1
		end
	end

	XPerl_SetFrameSizes()

	XPerl_Frame_ToggleLabels:SetButtonTex()

	return true
end

function XPerl_Toggle_ShowMT(noChange)
	if (not noChange) then
		if (XPerlConfigHelper.ShowMT == 1) then
			XPerlConfigHelper.ShowMT = 0
		else
			XPerlConfigHelper.ShowMT = 1
		end
	end

	XPerl_SetFrameSizes()

	XPerl_Frame_ToggleShowMT:SetButtonTex()

	return true
end

-- SmoothBarColor
local function SmoothBarColor(bar, barBG)
	local barmin, barmax = bar:GetMinMaxValues()
	local percentage = (bar:GetValue()/(barmax-barmin))

	local r, g
	if (percentage < 0.5) then
		r = 2*percentage
		g = 1
	else
		r = 1
		g = 2*(1 - percentage)
	end

	if ((r>=0) and (g>=0) and (r<=1) and (g<=1)) then
		bar:SetStatusBarColor(r, g, 0)
		barBG:SetVertexColor(r, g, 0, 0.25)
	end
end
